/*********************************************************************
 * Desktop Station shield 2 Sketch
 *
 * Copyright (C) 2021 DesktopStation
 *
 */

// include the library code:
#include <EEPROM.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_MM2.h"
#include "DSCoreM.h"
#include "Wire.h"
#include "Functions.h"
#include "TrackReporterS88_DS.h"


#define MAX_S88DECODER 1
#define RELPYERROR_300 "300 Command error"
#define RELPYERROR_301 "301 Syntax error"
#define RELPYERROR_302 "302 receive timeout"
#define RELPYERROR_303 "303 Unknown error"
#define RELPYERROR_NONE ""

#define ENABLE_RAILCOM 1	//0: Disable, 1: Enable

#define LEDSTATE_STOP 0
#define LEDSTATE_RUN 1
#define LEDSTATE_ERR 2

#define PIN_RUNLED	2


#define THRESHOLD_EDC 375  // 1024d * (1.2 / 5.9) * 9V / 5V
#define THRESHOLD_OV  999  // 1024d * (1.2 / 5.9) * 20V / 5V

/* Decralation classes */
DSCoreLib DSCore;
TrackReporterS88_DS reporter(MAX_S88DECODER);

/* LED functions */
byte gLED_State = LEDSTATE_STOP;
byte gLED_Toggle = 0;
byte gLED_Counter = 0;

/* Parser */
String gRequest;
String function;
word arguments[4];
word numOfArguments;
byte numOfLocAddr = 0;
word gLocAddresses[4] = {0,0,0,0};
boolean result;

//Task Schedule
unsigned long gPreviousL5 = 0; //   50ms interval
unsigned long gPreviousL8 = 0; // 1000ms interval
unsigned long gPreviousL9 = 0; // 2000ms interval



void printString(const char *s, int x, int y);
void ReplyPowerPacket(byte inPower);

/***********************************************************************
* 
* Boot setup
* 
* 
************************************************************************/

void setup()
{
	/* Debug message for PC */
	Serial.begin(115200);
	while (!Serial);
	
	Serial.println("-------------------------------------");
	Serial.println("Desktop Station shield R2 2021/1/30  ");
	Serial.println("-------------------------------------");
	Serial.println("100 Ready");

	DSCore.Init();
	DSCore.gCutOut = ENABLE_RAILCOM;

	//Pin mode
	pinMode(A1, INPUT);//Edc
	pinMode(9, OUTPUT);//Edc
	pinMode(10, OUTPUT);//Edc
	
	/* LED turn on */
	changeLED();
	
	/* Clear data of the S88 decoder */
	reporter.refresh();
	
	
#if defined(__AVR_ATmega4809__)
	
	/*
	TCA1.SINGLE.PER = 256;

	// CMP sets the duty cycle of the PWM signal -> CT = CMP0 / PER
	// DUTY CYCLE is approximately 50% when CMP0 is PER / 2
	TCA1.SINGLE.CMP0 = 200;

	// Counter starts at 0
	TCA1.SINGLE.CNT = 0x00;

	// Configuring CTRLB register
	// Compare 0 Enabled: Output WO0 (PB0) is enabled
	// Single slope PWM mode is selected
	TCA1.SINGLE.CTRLB = TCA_SINGLE_CMP0EN_bm | TCA_SINGLE_WGMODE_SINGLESLOPE_gc;

	// Using system clock (no frequency division, the timer clock frequency is Fclk_per)
	// Enable the timer peripheral
	TCA1.SINGLE.CTRLA = TCA_SINGLE_CLKSEL_DIV1_gc | TCA_SINGLE_ENABLE_bm;
	*/

#else
	//Timer1 490Hz -> 31.4kHz
	TCCR1B &= B11111000;
	TCCR1B |= B00000001;	
#endif
	
	//Reset task
	gPreviousL5 = millis();
	gPreviousL8 = millis();
	
	
	sei();    //ENABLE INTERRUPTION
	
}


boolean parse() 
{
	int aPosSlash = -1;

	int lpar = gRequest.indexOf('(');
	if (lpar == -1) 
	{
		return false;
	}

	function = String(gRequest.substring(0, lpar));
	function.trim();

	int offset = lpar + 1;
	int aAdddrOffset = 0;
	int comma = gRequest.indexOf(',', offset);
	numOfArguments = 0;
	numOfLocAddr = 0;
	
	while (comma != -1) 
	{
		String tmp = gRequest.substring(offset, comma);
		tmp.trim();
		
		/* スラッシュを探索する */
		aPosSlash = tmp.indexOf('/', 0);
		
		if( aPosSlash > 0)
		{
			tmp = tmp + "/";
			
			//Dammy値を指定
			arguments[numOfArguments++] = 0;
			aAdddrOffset = 0;
			
			//スラッシュはアドレス部だけに許可
			while( aPosSlash != -1)
			{
				String tmp2 = tmp.substring(aAdddrOffset, aPosSlash);
				tmp2.trim();
				
				if( numOfLocAddr < 4)
				{
					gLocAddresses[numOfLocAddr] = stringToWord(tmp2);
					numOfLocAddr++;
				}
				
				aAdddrOffset = aPosSlash + 1;
				aPosSlash = tmp.indexOf('/', aAdddrOffset);
			}
			
			offset = comma + 1;
			comma = gRequest.indexOf(',', offset);
		}
		else
		{
			arguments[numOfArguments++] = stringToWord(tmp);
			offset = comma + 1;
			comma = gRequest.indexOf(',', offset);
		}
	}

	int rpar = gRequest.indexOf(')', offset);
	while (rpar == -1) 
	{
		return false;
	}

	if (rpar > offset) 
	{
		String tmp = gRequest.substring(offset, rpar);
		tmp.trim();
		arguments[numOfArguments++] = stringToWord(tmp);
	}

	return true;
}

boolean dispatch() 
{
	byte aValue;
	byte aValueHigh;
	boolean aResult;
	char aTempStr[16];
	word aLocProtocol;
	byte aDir = 0;
	long aSpeed = 0;
	
	if ((function == "setLocoDirection") || ((function == "sLD")))
	{
		
		if( arguments[1] > 0)
		{
			//FWD=1,REV=2を FWD=0,REV=1に変換
			aDir = arguments[1] - 1;
			
			//異常値はREV固定
			if(aDir > 1)
			{
				aDir = 1;
			}
		}
		else
		{
			//FWD固定
			aDir = 0;
		}
		
		return DSCore.SetLocoDirection(arguments[0], aDir);
	
	}
	else if ((function == "setLocoFunction") || ((function == "sLF")))
	{
		return DSCore.SetLocoFunction(arguments[0], arguments[1] + 1, arguments[2]);
	
	}
	else if ((function == "setTurnout") || ((function == "sTO")))
	{
		return DSCore.SetTurnout(arguments[0], (byte)arguments[1]);
	
	}
	else if ((function == "setPower") || ((function == "sPW")))
	{
		
		if( arguments[0] == 1)
		{
			/* 状態変更 */
			gLED_State = LEDSTATE_RUN;
		}
		else
		{
			/* 状態変更 */
			gLED_State = LEDSTATE_STOP;
		}
		
		
		/* LED状態の変更 */
		changeLED();
		
		return DSCore.SetPower(arguments[0]);
	
	}
	else if ((function == "setLocoSpeed") || ((function == "sLS")))
	{
		return DSCore.SetLocoSpeedEx(arguments[0], arguments[1], arguments[2]);
	}
	else if ((function == "reset") || ((function == "sRS")))
	{
		Serial.println("100 Ready");
		return true;
	}
	/* reset */
	else if ((function == "setPing") || ((function == "sPI")))
	{
		/* Type(2bytes), Version(2bytes), UID(4bytes) */
		Serial.println("@DSG,00,03,00,02,00,00,00,00");
		return true;
	}
	else if ((function == "getLocoConfig") || ((function == "gCV")))
	{
		/* LED turn on */
		digitalWrite(PIN_RUNLED, HIGH);


		aValue = 0;
		aResult = DSCore.ReadConfig( arguments[1], &aValue, 0, 0);
		
		
		//Send to Serial
		Serial.print("@CV,");
		Serial.print(arguments[0]);
		Serial.print(",");
		Serial.print(arguments[1]);
		Serial.print(",");
		
		if( aResult == true)
		{
			Serial.print(aValue);
		}
		else
		{
			Serial.print(0xFFFF, HEX);
		}
		
		Serial.println(",");
		/* LED turn off */
		changeLED();		
		return true;
	}
	else if (function == "getS88")
	{
		int aMaxS88Num = MAX_S88DECODER;
		
		if( arguments[0] > 0)
		{
			aMaxS88Num = arguments[0];
		}
		
		//Send a S88 sensor reply 
		Serial.print("@S88,");

		reporter.refresh(aMaxS88Num);

		word aFlags = 0;

		for( int j = 0; j < aMaxS88Num; j++)
		{
		  aFlags = (reporter.getByte((j << 1) + 1) << 8) + reporter.getByte(j << 1);
		  
		  Serial.print((word)aFlags, HEX);
		  Serial.print(",");
		}
			
		Serial.println("");
		
		return true;
	} /* getS88 */
	else if ((function == "setLocoConfig") || ((function == "sCV")))
	{
		Serial.print("@CVWrite,");
		Serial.print(arguments[1]);
		Serial.print(",");
		Serial.print(arguments[2]);
		Serial.println("");

		/* LED turn on */
		digitalWrite(PIN_RUNLED, HIGH);

		DSCore.WriteConfig_Dir( arguments[1], arguments[2]);

		/* LED turn off */
		changeLED();

	}
	else if ((function == "cutout") || ((function == "CO")))
	{
		/* Railcom CutOut */
		DSCore.gCutOut = (arguments[0] == 1) ? 1: 0;
		
		return true;
	}

	else
	{
		return false;
	}
}

void loop()
{
	uint8_t aErrNo = 0;
	word aEdc;
	
	while(1)
	{
		aErrNo = 0;
		
		/* Reset watchdog timer. */
		//wdt_reset(); 
		
		//パルスジェネレータのスキャン
		DSCore.Scan();	
		
		char *aReplyText = RELPYERROR_NONE;
		
		if( (millis() - gPreviousL8) >= 200)
		{
			
			//LED Control
			taskLED();
			
			/* Sensor check */
			aEdc = analogRead(A1);
			
			/* OV check */
			if( aEdc > THRESHOLD_OV )
			{
				aErrNo = 2;
			}
			/* LV check */
			if( aEdc < THRESHOLD_EDC )
			{
				aErrNo = 1;
				
			}		
			/* エラー処理 */
			if( aErrNo > 0)
			{
				if( DSCore.IsPower() == 1)
				{
					PowerOffByErr(aErrNo);
				}
				
			}
		
			//Reset task
			gPreviousL8 = millis();
			
		}
		
		int aReceived = receiveRequest();
		
		
		
		if( aReceived > 0)
		{
  			#ifdef DEBUG
			Serial.println(gRequest);
			#endif
			
	  		if (parse()) {
				if (dispatch()) {
					aReplyText = "200 Ok";
				} else {
					//Serial.println(function);
					aReplyText = RELPYERROR_300;
				}
	  			
			} else {
				aReplyText = RELPYERROR_301;
			}
			
			/* Reply to Desktop Station */
			Serial.println(aReplyText);
			
		}
		else
		{
			/* Nothing to do */
		}
	}
}

int receiveRequest() {
	char buffer[64];
	int i = 0;
	char aByte;
	int aResult = -1;
	
	unsigned long time = millis();
	
	/* Check serial buffer */
	if (Serial.available() > 0) {
		
		while (1) {
		
			if (Serial.available() > 0) {
			
				aByte = Serial.read();
				
				// Write to text buf.
				buffer[i] = aByte;
				
				// check last identification of text end.
				if (aByte == '\n') {
					aResult = i;
					break;
				}
				
				// increment
				i = i + 1;
				
				// buf over check
				if( i > 64)
				{
					break;
				}
			}
			else
			{
				/* Check timeout. */
				if( millis() > time + 300)
				{
					aResult = -1;
					break;
				}
			}
		}
	}
	else
	{
		/* Not received. */
		aResult = 0;
	}
	
	if( aResult > 0)
	{
	
		buffer[i] = '\0';
		
		gRequest = String(buffer);
	}
	else
	{
		gRequest = "";
	}
	
	return aResult;
}

void PowerOffByErr(uint8_t inErrNo)
{
	DSCore.SetPower(0);
	delay(1000);
	ReplyPowerPacket(0);
	
	/* 状態変更 */
	gLED_State = LEDSTATE_ERR;
}

void ReplyPowerPacket(byte inPower)
{
	/* Power */
	Serial.print("@PWR,");
	Serial.print(inPower, HEX);
	Serial.println(",");
}


void changeLED(void)
{
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		digitalWrite(PIN_RUNLED, LOW);
		break;
		
	case LEDSTATE_RUN:
		digitalWrite(PIN_RUNLED, HIGH);
		break;
		
	case LEDSTATE_ERR:
		break;

	}
}

void taskLED(void)
{
	/* カウンタを回す(200msで1カウント) */
	gLED_Counter++;
	gLED_Counter = gLED_Counter & 0b1111;
	
	//LED操作
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		
		if( gLED_Counter <= 3)
		{
			digitalWrite(PIN_RUNLED, HIGH);
		}
		else
		{
			digitalWrite(PIN_RUNLED, LOW);
		}
		break;
		
	case LEDSTATE_RUN:
		break;
		
	case LEDSTATE_ERR:
		
		if( gLED_Toggle == 0)
		{
			digitalWrite(PIN_RUNLED, LOW);
		}
		else
		{
			digitalWrite(PIN_RUNLED, HIGH);
		}
		
		gLED_Toggle = (~gLED_Toggle) & 1;
		
		break;
	}
}

